==> gtoolkit-0.9.0.tar.gz: gtoolkit-0.9.0/GToolKit/GTKAction.m <==
/*
 * GToolKit - Objective-C interface to the GIMP Toolkit
 * Copyright (c) 1998  Elmar Ludwig - Universitaet Osnabrueck
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <GToolKit/GTKAction.h>
#include <GToolKit/GTK.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

@implementation GTKAction

/*
 * Create an autoreleased GTKAction that will send /selector/ to /target/.
 * @see -performWithSender:args:count:
 */
+ actionWithTarget:_target selector:(SEL) _selector
{
    return [[[self alloc] initWithTarget:_target selector:_selector]
	    autorelease];
}

/*
 * Create an autoreleased GTKAction that will send /selector/ to /target/.
 * /data/ will be passed as an additional argument to the method.
 */
+ actionWithTarget:_target selector:(SEL) _selector data:(const void *) _data
{
    return [[[self alloc] initWithTarget:_target selector:_selector data:_data]
	    autorelease];
}

/*
 * Initialize a new GTKAction without /selector/ and /target/.
 * @see -setTarget:selector:
 */
- init
{
    return [self initWithTarget:nil selector:0];
}

/*
 * Initialize a new GTKAction that will send /selector/ to /target/.
 * @see -performWithSender:args:count:
 */
- initWithTarget:_target selector:(SEL) _selector
{
    return [self initWithTarget:_target selector:_selector data:nil];
}

/*
 * Initialize a new GTKAction that will send /selector/ to /target/.
 * /data/ will be passed as an additional argument to the method.
 */
- initWithTarget:_target selector:(SEL) _selector data:(const void *) _data
{
    [[super init] setTarget:_target selector:_selector];
    [self setData:_data];
    return self;
}

/*
 * Set the receiver's target and/or selector. If /target/ is |nil| or
 * /selector/ is |NULL|, the previous value is not changed.
 */
- (void) setTarget:_target selector:(SEL) _selector
{
    if (_target || _selector == 0) target = _target;
    if (_selector) selector = _selector;

    if (selector)
    {
	if ([target respondsToSelector:selector])
	    method = [target methodForSelector:selector];
	else
	    [target doesNotRecognizeSelector:selector];
    }
}

/*
 * Set the receiver's user data.
 */
- (void) setData:(const void *) _data
{
    data = (void *) _data;
}

/*
 * Return the current target of this action.
 */
- target
{
    return target;
}

/*
 * Return the current selector of this action.
 */
- (SEL) selector
{
    return selector;
}

/*
 * Return the current user data of this action.
 */
- (void *) data
{
    return data;
}

/*
 * Send the current selector to the target. The action will pass the
 * /sender/ object, /count/ arguments from the argument array starting at
 * /args/ and the user data (if it has been set) to the method. The return
 * value will be stored in args[count], which must contain the correct type
 * information for the return value.<p>
 * If the selector's return type is |void| and /count/ is |0|,
 * args may be |NULL|.
 * @see GTK#-connectSignal:withAction: in class @GTK
 */
- (void) performWithSender:sender args:(GtkArg *) args count:(unsigned) count
{
#define MAXARGS	3
    int (*func)(id, SEL, ...) = (int (*)(id, SEL, ...)) method;
    int is_int[MAXARGS];	// is this type "compatible" to int?
    int num[MAXARGS];		// integer value of the argument
    void *ptr[MAXARGS];		// pointer value of the argument
    int index;
    int result;

    // this code currently does not use the NSInvocation class because:
    //	 - NSInvocation is implemented only on a few systems (yet)
    //	 - NSInvocation is relatively slow

    // allow up to MAXARGS arguments of type INT, LONG, FLOAT or POINTER
    for (index = 0; index != MAXARGS; ++index)
    {
	if (index < count)
	{
	    GtkType type = args[index].type;

#if SIZEOF_LONG != SIZEOF_INT && SIZEOF_LONG != SIZEOF_VOID_P
	    if (type == GTK_TYPE_LONG || type == GTK_TYPE_ULONG)
		g_warning("cannot pass GTK_TYPE_LONG on this machine");
#endif
#if SIZEOF_FLOAT != SIZEOF_INT && SIZEOF_FLOAT != SIZEOF_VOID_P
	    if (type == GTK_TYPE_FLOAT)
		g_warning("cannot pass GTK_TYPE_FLOAT on this machine");
#endif
	    is_int[index] = type == GTK_TYPE_INT   || type == GTK_TYPE_UINT  ||
#if SIZEOF_INT == SIZEOF_LONG
			    type == GTK_TYPE_LONG  || type == GTK_TYPE_ULONG ||
#endif
			    type == GTK_TYPE_BOOL  || type == GTK_TYPE_ENUM  ||
#if SIZEOF_INT == SIZEOF_FLOAT
			    type == GTK_TYPE_FLOAT ||
#endif
			    type == GTK_TYPE_FLAGS;
	    num[index] = GTK_VALUE_INT(args[index]);
	    ptr[index] = gtk_type_is_a(type, GTK_TYPE_OBJECT) ?
			    Gtk_to_Object(GTK_VALUE_OBJECT(args[index]), 0) :
			    GTK_VALUE_POINTER(args[index]);
	}
	else
	{
	    is_int[index] = 0, ptr[index] = data;
	    break;
	}
    }

    // method is always called with the maximum number of arguments,
    // this is ok because Objective-C uses the C calling convention
    result =
    (is_int[0] ?
	(is_int[1] ?
	    (is_int[2] ?
		func(target, selector, sender, num[0], num[1], num[2], data)
	     :
		func(target, selector, sender, num[0], num[1], ptr[2], data)
	    )
	 :
	    (is_int[2] ?
		func(target, selector, sender, num[0], ptr[1], num[2], data)
	     :
		func(target, selector, sender, num[0], ptr[1], ptr[2], data)
	    )
	)
     :
	(is_int[1] ?
	    (is_int[2] ?
		func(target, selector, sender, ptr[0], num[1], num[2], data)
	     :
		func(target, selector, sender, ptr[0], num[1], ptr[2], data)
	    )
	 :
	    (is_int[2] ?
		func(target, selector, sender, ptr[0], ptr[1], num[2], data)
	     :
		func(target, selector, sender, ptr[0], ptr[1], ptr[2], data)
	    )
	)
    );

    // return type may be VOID or INT
    if (args && args[count].type != GTK_TYPE_NONE)
	*GTK_RETLOC_INT(args[count]) = result;
#undef MAXARGS
}
@end

==> gtoolkit-0.9.2.tar.gz: gtoolkit-0.9.2/GToolKit/GTKAction.m <==
/*
 * GToolKit - Objective-C interface to the GIMP Toolkit
 * Copyright (c) 2000  Elmar Ludwig - Universitaet Osnabrueck
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <GToolKit/GTKAction.h>
#include <GToolKit/GTK.h>

@implementation GTKAction

/*
 * Create an autoreleased GTKAction that will send /selector/ to /target/.
 * @see -performWithSender:args:count:
 */
+ actionWithTarget:_target selector:(SEL) _selector
{
    return [[[self alloc] initWithTarget:_target selector:_selector]
	    autorelease];
}

/*
 * Create an autoreleased GTKAction that will send /selector/ to /target/.
 * /data/ will be passed as an additional argument to the method.
 * Note that /data/ is not retained by default.
 */
+ actionWithTarget:_target selector:(SEL) _selector data:(const void *) _data
{
    return [[[self alloc] initWithTarget:_target selector:_selector data:_data]
	    autorelease];
}

/*
 * Initialize a new GTKAction without /selector/ and /target/.
 * @see -setTarget:selector:
 */
- init
{
    return [self initWithTarget:nil selector:0];
}

/*
 * Initialize a new GTKAction that will send /selector/ to /target/.
 * @see -performWithSender:args:count:
 */
- initWithTarget:_target selector:(SEL) _selector
{
    return [self initWithTarget:_target selector:_selector data:nil];
}

/*
 * Initialize a new GTKAction that will send /selector/ to /target/.
 * /data/ will be passed as an additional argument to the method.
 * Note that /data/ is not retained by default.
 */
- initWithTarget:_target selector:(SEL) _selector data:(const void *) _data
{
    [[super init] setTarget:_target selector:_selector];
    [self setData:_data];
    return self;
}

 - (void) dealloc
{
    if (retained) [(id) data release];
    [super dealloc];
}

/*
 * Set the receiver's target and/or selector. If /target/ is |nil| or
 * /selector/ is |NULL|, the previous value is not changed.
 */
- (void) setTarget:_target selector:(SEL) _selector
{
    if (_target || _selector == 0) target = _target;
    if (_selector) selector = _selector;

    if (selector)
    {
	if ([target respondsToSelector:selector])
	    method = [target methodForSelector:selector];
	else
	    [target doesNotRecognizeSelector:selector];
    }
}

/*
 * Set the receiver's user data.
 */
- (void) setData:(const void *) _data
{
    if (retained) [(id) _data retain], [(id) data release];
    data = (void *) _data;
}

/*
 * Tell the receiver to retain its user data. Use this only if the data
 * really is some kind of NSObject.
 */
- (void) retainData
{
    if (retained == NO) [(id) data retain], retained = YES;
}

/*
 * Return the current target of this action.
 */
- target
{
    return target;
}

/*
 * Return the current selector of this action.
 */
- (SEL) selector
{
    return selector;
}

/*
 * Return the current user data of this action.
 */
- (void *) data
{
    return data;
}

/*
 * Return whether the user data should be retained.
 */
- (BOOL) dataRetained
{
    return retained;
}

/*
 * Send the current selector to the target. The action will pass the
 * /sender/ object, /count/ arguments from the argument array starting at
 * /args/ and the user data (if it has been set) to the method. The return
 * value will be stored in args[count], which must contain the correct type
 * information for the return value.<p>
 * If the selector's return type is |void| and /count/ is |0|,
 * args may be |NULL|.
 * @see GTK#-connectSignal:withAction: in class @GTK
 */
- (void) performWithSender:sender args:(GtkArg *) args count:(unsigned) count
{
#define MAXARGS	3
    int (*func)(id, SEL, ...) = (int (*)(id, SEL, ...)) method;
    int is_int[MAXARGS];	// is this type "compatible" to int?
    int num[MAXARGS];		// integer value of the argument
    void *ptr[MAXARGS];		// pointer value of the argument
    int index;
    int result;

    // this code currently does not use the NSInvocation class because:
    //	 - NSInvocation is implemented only on a few systems (yet)
    //	 - NSInvocation is relatively slow

    // allow up to MAXARGS arguments of type INT, LONG, FLOAT or POINTER
    for (index = 0; index != MAXARGS; ++index)
    {
	if (index < count)
	{
	    GtkType type = args[index].type;

#ifdef DEBUG
	    if (type == GTK_TYPE_NONE || type == GTK_TYPE_INVALID)
		g_warning("cannot pass GTK_TYPE_NONE as argument");
#endif
#if SIZEOF_LONG != SIZEOF_INT && SIZEOF_LONG != SIZEOF_VOID_P
	    if (type == GTK_TYPE_LONG || type == GTK_TYPE_ULONG)
		g_warning("cannot pass GTK_TYPE_LONG on this machine");
#endif
#if SIZEOF_FLOAT != SIZEOF_INT && SIZEOF_FLOAT != SIZEOF_VOID_P
	    if (type == GTK_TYPE_FLOAT)
		g_warning("cannot pass GTK_TYPE_FLOAT on this machine");
#endif
	    is_int[index] = type == GTK_TYPE_INT   || type == GTK_TYPE_UINT  ||
#if SIZEOF_INT == SIZEOF_LONG
			    type == GTK_TYPE_LONG  || type == GTK_TYPE_ULONG ||
#endif
			    type == GTK_TYPE_BOOL  || type == GTK_TYPE_ENUM  ||
#if SIZEOF_INT == SIZEOF_FLOAT
			    type == GTK_TYPE_FLOAT ||
#endif
			    type == GTK_TYPE_FLAGS;
	    num[index] = GTK_VALUE_INT(args[index]);
	    ptr[index] = gtk_type_is_a(type, GTK_TYPE_OBJECT) ?
			    GTK_GET_OBJECT(GTK_VALUE_OBJECT(args[index])) :
			    GTK_VALUE_POINTER(args[index]);
	}
	else
	{
	    is_int[index] = 0, ptr[index] = data;
	    break;
	}
    }

    // method is always called with the maximum number of arguments,
    // this is ok because Objective-C uses the C calling convention
    result =
    (is_int[0] ?
	(is_int[1] ?
	    (is_int[2] ?
		func(target, selector, sender, num[0], num[1], num[2], data)
	     :
		func(target, selector, sender, num[0], num[1], ptr[2], data)
	    )
	 :
	    (is_int[2] ?
		func(target, selector, sender, num[0], ptr[1], num[2], data)
	     :
		func(target, selector, sender, num[0], ptr[1], ptr[2], data)
	    )
	)
     :
	(is_int[1] ?
	    (is_int[2] ?
		func(target, selector, sender, ptr[0], num[1], num[2], data)
	     :
		func(target, selector, sender, ptr[0], num[1], ptr[2], data)
	    )
	 :
	    (is_int[2] ?
		func(target, selector, sender, ptr[0], ptr[1], num[2], data)
	     :
		func(target, selector, sender, ptr[0], ptr[1], ptr[2], data)
	    )
	)
    );

#ifdef DEBUG
    if (args && args[count].type == GTK_TYPE_INVALID)
	g_warning("cannot return GTK_TYPE_INVALID");
#endif
    // return type may be VOID or INT (or "compatible")
    if (args && args[count].type != GTK_TYPE_NONE)
	*GTK_RETLOC_INT(args[count]) = result;
#undef MAXARGS
}
@end
